/*----------------------------------------------------------------------------*
 *                                                                            *
 *   HPF Array Combining Scatter Function                                     *
 *                                                                            *
 *     - Author: Seungjo Bae & Khaled A. Alsabti                              *
 *                                                                            *
 *----------------------------------------------------------------------------*/

/******************************************************************************

  History

     1)  Make portable as much as possible 
         Do not share any variables with other routines (6/15/97)

     2)  Combine BLOCK with BLOCK-CYCLIC in PLC (6/26/97)

     3)  Combine PLC with NLC (6/27/97)

     4)  Two version: Separate & Combine (combine all routines into one)
         (7/1/97)

     5)  Data_Type instead of int

     6)  New buffer type: Struct Buffer_Type = data + ptr (7/9/97)

     7)  Remove mask... and shift (7/9/97)

     8)  Generalize comm. scheme in order to change it to MPI (7/9/97)
         File = "comm-raw.c"

     9)  New interface (7/12/97)

     10) General version 1: No assumption for N, N', P, but K = 1

          - Fixed MP_div_K & MPS_div_K (7/12/97)
            That is, fixed bucket size for all target processors

          - Fixed NP_div_K & NPS_div_K (7/12/97)

          - refer to / *--- GENERAL ---* /

     11) Make  hot processor 0 the last processor with local elements
         instead of P0 (7/12/97)

     12) Use "MP_div_K_times_K" instead of "MP_max" when computing prank
          (7/13/97)

         ===> Now no assumption for K

     13) Works when s_tot_nproc != t_tot_nproc.

         However we need carefule examination for optimization,
         that is, how to decide if a bucket is globally dense or sparse 
         (7/13/97)

         ===> Now, complete general version for N, N', P, P', K

     14) Change MP_div_K_times_K to cirtual_MP
                MPS_div_K        to locally_dense_limit (7/14/97)

     15) Take care of problems in (13)
           - Introduce globally_dense_limit (7/14/97)

     16) Optimiza the last message decomposition in the second stage (7/15/97)

     -----

     17) First MPI version: Fixed nproc (7/23/97)

     18) MPI: Various nproc by generating new communicator "WORLD" (7/23/97)

     19) Bug in dealing with s_tot_nproc > t_tot_nproc
         ===> Change t_tot_nproc to tot_nproc when dealing with comm. vectors
              Why? It is not guaranteed that each vector element is 
              initialized to 0.
              (7/24/97)

     20) Test Data_Type = double. It's work. (7/25/97)

     21) Use "source_lidx"  (7/25/97)

     22) Break routines into two parts: (8/8/97)
            I.  Common parts for all data types
	    II. Dealing with source or target arrays or buffers
	        i.e., dependent on the data type

     23) Combine CM5 with MPI version (8/8/97)

     24) This file only contains the common part (8/9/97)

     -----

     25) Interface with Bryan (8/27/97)

         - New Interface

	     void nplc_tsc(
	     
	     #if CM5
	     #else
	       MPI_Comm  WORLD_input,
	     #endif
	     
	       void *source_array,         
	       int source_lidx[],
	       int NP,                       / * s_L_all * /             
	       int s_N_all_input,            / * s_N_all * / 
					     
	       int index_lidx[],
	       int index_pnum[],   
	      
	       void *target_array,     
	       int MP,                      / * t_L_all * /                     
					    / * All proc. have the same MP * /
	       Data_Types data_type_input,
	       Operator_Types operator_type_input,
	     
	       int K_input,
	       int PLC_input,
	       double local_dense_factor_input,
	       int algorithm)


        - MP is uniform for all target processors

	- s_tot_nproc = t_tot_nproc

     26) Make some global variables local (9/9/97)

 ******************************************************************************/

  
#include "raw_global.h"

/******************************************************************************
 *                                                                            *
 *  Function : 
 *                                                                            *
 ******************************************************************************/

void nplc_tsc(

#if CM5
#else
	      MPI_Comm  WORLD_input,
#endif

	      void *source_array,         
	      int source_lidx[],
	      int NP,                       /* s_L_all */             
	      int s_N_all_input,            /* s_N_all */ 
				
	      int index_lidx[],
	      int index_pnum[],   
 
	      void *target_array,     
	      int MP,                      /* t_L_all */                      
	                                   /* All proc. have the same MP */
	      Data_Types data_type_input,
	      Operator_Types operator_type_input

	      /***------------------------***
	      ,
	      int K_input,
	      int PLC_input,
	      double local_dense_factor_input,
	      int algorithm_input
	      ***------------------------***/
	      )

{

  void INT_nplc_tsc();
  void FLOAT_nplc_tsc();
  void DOUBLE_nplc_tsc();
  void LOGIC_nplc_tsc();
  void COMPLEX_nplc_tsc();

  int i;
  int algorithm;

  double local_dense_factor;
  double global_dense_factor;

  /*---------------------------- Proc. Info ---------------------------------*/

#if CM5		
  tot_nproc = CMMD_partition_size();
  my_pnum   = CMMD_self_address(); 

#else
  WORLD = WORLD_input;
  MPI_Comm_size(WORLD,&tot_nproc);
  MPI_Comm_rank(WORLD,&my_pnum);

#endif

  s_tot_nproc = t_tot_nproc = tot_nproc;

  /*------------------------ Setup global variables  ------------------------*/
  
  /*** s_N_all = s_N_all_input; ***/

  /***---    From arguments    ---***
  if ( (K_input<1) || (K_input>MP) ) 
    K = Default_K;
  else
    K = K_input;

  PLC = PLC_input;

  if ( local_dense_factor_input < 1. ) 
    local_dense_factor = Default_local_dense_factor;
  else
    local_dense_factor  = local_dense_factor_input; 

  global_dense_factor = Default_global_dense_factor;

  if ( algorithm_input < 0 )
    algorithm = Default_algorithm;
  else
    algorithm = algorithm_input;
  ***-----------------------------***/

  /***---  From default values  ---***/
  K = Default_K;
  PLC = 0;
  local_dense_factor = Default_local_dense_factor;
  global_dense_factor = Default_global_dense_factor;
  algorithm = Default_algorithm;
  /***-----------------------------***/

  /*------------------- determining the number of buckets  -------------------*/

  KPS = K * t_tot_nproc;

  if ( KPS > MAXINT ) {
    printf("\n Number of bucket %d is too large. Exit! \n\n",KPS);
    exit(0);
  }

  /*--- GENERAL ---*/

  /* Information for target bucket */

  MP_div_K = Ceiling(MP, K);          /* fixed target bucket size */
  virtual_MP = MP_div_K * K;          /* virtual local size at target proc */
                                      /* Used for computing prank */
                                      /* virtual_MP >= MP */ 

  locally_dense_limit = MP_div_K * local_dense_factor; 
                                      /* locally dense or sparse */

  /* For determining globally dense or sparse                          */
  /* Aver # of source elements per target bucket X global_dense_factor */
  globally_dense_limit = ( s_N_all_input / KPS ) * global_dense_factor;       
  
  /*----------------------------- Allocation  -------------------------------*/

  cache_B = (int *) malloc(NP * sizeof(int));

  rdispls = (int *) malloc(tot_nproc*sizeof(int));
  recvcount = (int *) malloc(tot_nproc*sizeof(int));
  sdispls = (int *) malloc(tot_nproc*sizeof(int));
  sendcount = (int *) malloc(tot_nproc*sizeof(int));
  ptr_bucket = (int *) malloc(KPS*sizeof(int));
  prefix_sum_count_bucket = (int *) malloc(KPS*sizeof(int));
  sum_count_bucket = (int *) malloc(KPS*sizeof(int));
  prefix_count_bucket = (int *) malloc(KPS*sizeof(int));
  count_bucket = (int *) malloc(KPS*sizeof(int));
  rdisplsSP = (int *) malloc(tot_nproc*sizeof(int));
  recvcountSP = (int *) malloc(tot_nproc*sizeof(int));
  sdisplsSP = (int *) malloc(tot_nproc*sizeof(int));
  sendcountSP = (int *) malloc(tot_nproc*sizeof(int));
  Sparse = (int *) malloc(KPS*sizeof(int));
  bucket_Start = (int *) malloc(KPS*sizeof(int));
 
  if  (/*cache_B           	== NULL ||*/
       rdispls            	== NULL ||
       recvcount       	  	== NULL ||
       sdispls		  	== NULL ||
       sendcount          	== NULL ||
       ptr_bucket 	      	== NULL ||
       prefix_sum_count_bucket 	== NULL ||
       sum_count_bucket	  	== NULL ||
       prefix_count_bucket 	== NULL ||
       count_bucket 		== NULL ||
       rdisplsSP 		== NULL ||
       recvcountSP		== NULL ||
       sdisplsSP		== NULL ||
       sendcountSP		== NULL ||
       Sparse		        == NULL ||
       bucket_Start		== NULL )  {
    printf(" Out of Memory: ....... in nlc_raw_setup()\n"); 
    exit(0); 

  }	

  DATA_TYPE = data_type_input;
  OP_TYPE   = operator_type_input;


  /*------------------------ Initialize count_bucket -------------------------*/

  for (i=0; i < KPS ; i++)
       count_bucket[i] = 0;

  /*------------------ Call a proper function for each type ------------------*/

  switch (DATA_TYPE) {

  case TYPE_int: 

    INT_nplc_tsc(source_array, source_lidx, NP,
		 index_lidx, index_pnum, 
		 target_array, MP,
		 algorithm);     
    break;

  case TYPE_float:

    FLOAT_nplc_tsc(source_array, source_lidx, NP,
		   index_lidx, index_pnum, 
		   target_array, MP, 
		   algorithm);     
    break;

  case TYPE_double:

    DOUBLE_nplc_tsc(source_array, source_lidx, NP, 
		    index_lidx, index_pnum, 
		    target_array, MP,
		    algorithm);     
    break;

  case TYPE_logic:

    LOGIC_nplc_tsc(source_array, source_lidx, NP, 
		   index_lidx, index_pnum, 
		   target_array, MP, 
		   algorithm);     
    break;

  case TYPE_complex:

    COMPLEX_nplc_tsc(source_array, source_lidx, NP, 
		     index_lidx, index_pnum, 
		     target_array, MP, 
		     algorithm);     
    break;

  }


  /*------------------------------ House Keeping -----------------------------*/

  if (count_bucket != NULL)               free(count_bucket); 
  if (sum_count_bucket != NULL)           free(sum_count_bucket); 
  if (prefix_sum_count_bucket != NULL)    free(prefix_sum_count_bucket); 
  if (ptr_bucket != NULL)                 free(ptr_bucket); 
  if (prefix_count_bucket != NULL)        free(prefix_count_bucket); 
                                          
  if (Sparse != NULL)                     free(Sparse);
  if (bucket_Start != NULL)               free(bucket_Start); 
                                        
  if (rdispls != NULL)                    free(rdispls); 
  if (recvcount != NULL)                  free(recvcount);
  if (sdispls != NULL)                    free(sdispls); 
  if (sendcount != NULL)                  free(sendcount);
  if (rdisplsSP != NULL)                  free(rdisplsSP); 
  if (recvcountSP != NULL)                free(recvcountSP);
  if (sdisplsSP != NULL)                  free(sdisplsSP); 
  if (sendcountSP != NULL)                free(sendcountSP);
                                        
/*** Cause segmentation fault in raw-2d, thus, in case delete it (3/2/97)
  if (proc_list != NULL)                  free(proc_list);
***/

}

/******************************************************************************
 *                                                                            *
 *  Function : stage1_pre()
 *                                                                            *
 ******************************************************************************/

void stage1_pre() 

{

  int pn, lidx;
  int ind, dest, size;
  int first, last;
  int i, j;

  /*=========================================================================*
   *===                     B E G I N   O F   P L C                       ===*   
   *=========================================================================*/

if (PLC) {

  LSparse = (int *) malloc((KPS*sizeof(int)));
  if  (LSparse== NULL) {
    printf(" Out of Memory: ....... in plc_raw_scan()\n");
    exit(0);
  }

  for (i=0 ; i < KPS ; i++ )
    if ( count_bucket[i] <= locally_dense_limit /**MPS_div_K**/ )
      LSparse[i] = 1;
    else  {
      LSparse[i] = 0;
      count_bucket[i] = MP_div_K;
    }
}

  /*=========================================================================*
   *===                        E N D   O F   P L C                        ===*   
   *=========================================================================*/

  /*--------- Performing a global vector sum-combine on count_bucket --------*/

#if CM5	

  CMMD_reduce_v(sum_count_bucket,count_bucket,
		CMMD_combiner_add,sizeof(int),KPS);
#else

  MPI_Allreduce( count_bucket, sum_count_bucket, KPS, 
		 MPI_INT, MPI_SUM, WORLD);
#endif

/*----------------------- Globally Sparse or Dense -------------------------*/

  for (i=0; i < tot_nproc ; i++)
     sendcountSP[i] = sendcount[i] = 0;

  ind = 0;
  for (i=0 ; i < KPS ; i++)

    /*--- GENERAL ---*/
    if ((sum_count_bucket[i] <= globally_dense_limit /**NPS_div_K**/ ))  { 
      Sparse[i] = 1; 

      if (count_bucket[i]) {
	bucket_Start[i] = ind;
	ind+= count_bucket[i];
	dest = i/K; 
	sendcountSP[dest] += count_bucket[i];
      }

      sum_count_bucket[i] = count_bucket[i] = 0; /* for resizing IP array */
    }
    else 
      Sparse[i] = 0;
 
  /*--------- Performing a global vector prefix-sum on count_bucket ----------*/

#if CM5	

  CMMD_scan_v(prefix_count_bucket,count_bucket,
	      CMMD_combiner_add,CMMD_upward,CMMD_none,0,
	      CMMD_exclusive,sizeof(int),KPS);
#else

  /* inclusive */
  MPI_Scan( count_bucket, prefix_count_bucket, KPS, 
	    MPI_INT, MPI_SUM, WORLD);
  for (i=0 ; i < KPS ; i++)	/* exclusive */
    prefix_count_bucket[i] -= count_bucket[i];

#endif


  /*-------- Performing a local execlusive prefix-sum on sum_count_bucket ----*/
  /*---------- Determining the new starting position of each bucket  ---------*/

  prefix_sum_count_bucket[0] = 0;

  for (i=0 ; i < KPS-1 ; i++) {
    ptr_bucket[i] = prefix_count_bucket[i] + prefix_sum_count_bucket[i];
    prefix_sum_count_bucket[i+1] = sum_count_bucket[i]+
                                prefix_sum_count_bucket[i];
  }
  ptr_bucket[i] = prefix_count_bucket[i] + prefix_sum_count_bucket[i];

  /*--------------------- Resizing intermediate arrays ----------------------*/

  IPsize_all = prefix_sum_count_bucket[KPS-1] + sum_count_bucket[KPS-1];
  IPsize = Ceiling(IPsize_all, t_tot_nproc);

  /*--------------------------- Update sendl vector -------------------------*/

  ind = 0;
  for (i=0; i < KPS ; i++)
    if ( (Sparse[i] ==0) && (count_bucket[i]) ) {
      bucket_Start[i] = ind;
      ind += count_bucket[i];
      first = ptr_bucket[i] / IPsize;
      last = (ptr_bucket[i]+count_bucket[i]-1) / IPsize;
      if ( first == last )  sendcount[first]+=count_bucket[i];
      else {
        sendcount[first]+= (IPsize-ptr_bucket[i]%IPsize);
        size = (ptr_bucket[i]+count_bucket[i])%IPsize;
        sendcount[last] += (size == 0) ? IPsize: size;
        for (j=first+1; j < last ; j++) 
	  sendcount[j] += IPsize; 
      }

    }

  /*----------------  Combine the receive vector, recvcount, -----------------*/

  Preprocessing_comm(sendcount,recvcount,sdispls,rdispls,
		     &total_send,&total_recv);

  Preprocessing_comm(sendcountSP,recvcountSP,sdisplsSP,rdisplsSP,
		     &total_sendSP,&total_recvSP);

}



/******************************************************************************
 *                                                                            *
 *  Function : stage2_pre
 *                                                                            *
 ******************************************************************************/

    /*=========================   N O T E   ================================*/
    /*===                                                                ===*/
    /*===     As the result of using "prank" instead of grank,           ===*/
    /*===     there is no difference between (Block) and                 ===*/
    /*===     (Block-Cyclic) in the second stage.                        ===*/
    /*===                                                                ===*/
    /*======================================================================*/

void stage2_pre() 

{
  int start, end, no_buckets, dest, i;
  int ind, first_full_bucket, last_full_bucket;

   /*--------- Determining The number of partial buckets,2,1 or 0 and the ----*/
   /*---------             total number of buckets                         ---*/

   no_buckets =0;
   start = my_pnum*IPsize;
   end = (my_pnum+1)*IPsize-1;
   ind = KPS-1;

   while ( (ind> 0 ) &&  (prefix_sum_count_bucket[ind] > start ))
     ind--;
   first_bucket = ind;
   while ( (ind <KPS) && (prefix_sum_count_bucket[ind] <= end )) {
     no_buckets++;
     ind++;
    }
   last_bucket = ind -1;


   /*---------------- determine if the first bucket is partial ---------------*/

   if ( (prefix_sum_count_bucket[first_bucket] < start ) &&
        ( Sparse[first_bucket] == 0) &&
/* ??*/ (prefix_sum_count_bucket[first_bucket]+sum_count_bucket[first_bucket] > start))
     first_bucket_partial = 1;  /* Yes */
   else
     first_bucket_partial = 0;  /* No */


   /*--------------- determine if the last bucket is partial -----------------*/

   if (((prefix_sum_count_bucket[last_bucket]+sum_count_bucket[last_bucket]-1)
      > end ) && (Sparse[last_bucket] == 0))
     last_bucket_partial = 1; /* Yes */
   else
     last_bucket_partial = 0; /* NO */


   /*--------------- Calculating the size of the sending buffer --------------*/

   if ( first_bucket_partial )  
     first_full_bucket = first_bucket + 1;
   else 
     first_full_bucket = first_bucket;
   if ( last_bucket_partial ) 
     last_full_bucket = last_bucket -1;
   else 
     last_full_bucket = last_bucket;
   if ( (last_bucket_partial)||(first_bucket_partial))
     Size = MP_div_K;
   else 
     Size = 0;
   if ( (last_bucket_partial)&&(first_bucket_partial)
       && ( first_bucket != last_bucket ))
     Size += MP_div_K;
   ind = 0;

   for (i=0; i < tot_nproc ; i++) {
     sendcount[i] = 0;
     sdispls[i] = -1;
     }

   for (i=first_full_bucket;  i <= last_full_bucket ; i++)
     if ( Sparse[i] == 0 ) {  /* dense bucket */
       ptr_bucket[i] = ind;
       dest = i/K;
       Size += MP_div_K;
       sendcount[dest] += MP_div_K;
       if ( sdispls[dest] == -1 ) sdispls[dest] = ind;
       ind += MP_div_K;
      }

   /*if ( (last_bucket_partial )&&((first_bucket_partial==0)||(
        last_bucket!=first_bucket)))   {*/

   /*--- My modification ---*/
   if ( (last_bucket_partial ) &&
        ( (last_bucket != first_bucket) || 
          (prefix_sum_count_bucket[last_bucket] == start) ) ) {
     ptr_bucket[last_bucket] = ind;
     dest = last_bucket/K;
     sendcount[dest] += MP_div_K;
     if ( sdispls[dest] == -1 )  sdispls[dest]= ind;
     ind += MP_div_K;
   }

   if  (first_bucket_partial)
      ptr_bucket[first_bucket] = ind;

}



/******************************************************************************
 *                                                                            *
 *  Function : stage2_pre_segment
 *                                                                            *
 ******************************************************************************/

void stage2_pre_segment() 

{
  int ind, i, j;
  int first_proc, last_proc, end_bucket_s, end_s;


  if ( (first_bucket_partial) &&
      ( (last_bucket_partial==0) || (first_bucket == last_bucket)) )
    BN = first_bucket;
  else if ( last_bucket_partial )  
    BN = last_bucket;
  else 
    BN = -1;
  sourcePN = -1;

  if ( BN != -1 ) {  /*  I am invloved in a segemented scan */
    
    /*-- determine the processors invloving in the segemented scan --*/

    first_proc = prefix_sum_count_bucket[BN] / IPsize;
    last_proc = (prefix_sum_count_bucket[BN]+sum_count_bucket[BN]-1)/IPsize;
    no_proc = last_proc - first_proc + 1;

    /* Check if last_proc has succeeding partial bucket, if so then
       delete it from the list */
    /* determine if last proc  has succeeding partial bucket */

    end_s = ( last_proc + 1 ) * IPsize - 1;
    ind = BN;
    while ( (ind < KPS ) && ( prefix_sum_count_bucket[ind] <= end_s) )
      ind++;
    end_bucket_s = ind -1;

    if ( ( (prefix_sum_count_bucket[end_bucket_s] +
            sum_count_bucket[end_bucket_s] -1 ) > end_s ) &&
	( Sparse[end_bucket_s] == 0 ) &&
	( BN != end_bucket_s ) )  {
      no_proc--;  
      if ( my_pnum == first_proc ) 
	sourcePN = last_proc;
      last_proc--;
    } 

    proc_list = (int *) malloc(no_proc*sizeof(int));
    if ( no_proc > 0 && proc_list == NULL ) {
      printf(" Out of Memory: proc_list in raw_stage2()\n"); 
      exit(0);
    }		

    for (i=0 ; i < no_proc ; i++)
      proc_list[i] = first_proc + i;

    /***
    if ( no_proc > 1 )

      Segmented_Sum_V(&sendbuf2[ptr_bucket[BN]],MP_div_K,proc_list,no_proc);
    ***/

  } /* if ( BN != -1 ) */
}



/*============================================*
 *  Integer Function : Clog2(n)               *
 *                                            *
 *   - Input  = positive integer              *
 *   - Return = Ceiling ( log2(n) )           *
 *============================================*/

int CLOG2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1<<i != n )
    return i+1;
  else
    return i;
}
